from typing import TypeVar, Callable, cast
from typing_extensions import ParamSpec

__all__ = [
    "F", "MISSING", "T"
]


class _MISSING:
    pass


MISSING = _MISSING()
F = TypeVar("F", bound=Callable)
F2 = TypeVar("F2", bound=Callable)
P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")


def cast_to_func(fn: F2, target: F) -> F:
    """
    Casts function ``fn`` to ``target`` to aid with autocomplete and whatnot.
    """
    return cast(F, fn)



from dataclasses import dataclass as dc
from functools import partial
from typing import Callable, Optional, Type, cast

from .Classes import classtypes
from .types import T, cast_to_func

__all__ = [
    "dataclass"
]


def _dataclass(cls: Optional[Type[T]] = None, **kwargs: bool) -> Callable[[Type[T]], Type[T]]:
    """
    Creates a type-safe dataclass, used just like @dataclasses.dataclass.
    """
    if cls is not None:
        return classtypes(dc(cls, **kwargs))  # type: ignore
    else:
        return cast(Callable, partial(_dataclass, **kwargs))


dataclass = cast_to_func(_dataclass, dc)



from functools import wraps
from inspect import _ParameterKind as PK
from inspect import signature
from typing import Any, cast
from typing_extensions import get_args

from .types import MISSING, F


__all__ = [
    "functypes"
]


def functypes(func: F) -> F:
    """For use as a function decorator, enforces types in ``__call__``"""
    assert func

    @wraps(func)
    def newcall(*args: Any, **kwargs: Any) -> Any:
        keywords = signature(func).parameters
        arglist = list(args)
        for kw in keywords:
            argval = arglist.pop(0) if len(arglist) > 0 else ...
            kw_val = kwargs.get(kw, MISSING)
            if argval is not ... and kw_val is not MISSING:
                raise SyntaxError(f"{kw} was given as both a positional and a keyword argument! "
                                  f"({func.__name__})")
            if kw_val is MISSING:
                if argval is ...:
                    continue
                kw_val = argval
                if keywords[kw].kind == PK.POSITIONAL_ONLY:
                    arglist.append(kw_val)
                else:
                    if keywords[kw].kind == PK.KEYWORD_ONLY and kw not in kwargs:
                        raise SyntaxError(f"{kw} is a keyword-only argument, but was given as "
                                          f"positional argument! ({func.__name__})")
                    kwargs[kw] = kw_val
            argtype = keywords[kw].annotation
            argtuple = get_args(argtype)
            if not isinstance(kw_val, argtype) and not isinstance(kw_val, argtuple):
                raise TypeError(f"Argument {kw} was passed a value of type `"
                                f"{type(kw_val).__name__}`, but only accepts values of "
                                f"type `{argtype.__name__}` "
                                f"({func.__name__})")
        return func(*arglist, **kwargs)

    return cast(F, newcall)



"""Simple decorators for enforcing types during runtime."""
# Flake8: noqa
from .Classes import classtypes, methtypes
from .Combined import dataclass
from .Functions import functypes

__version__ = "0.0.5"
__all__ = [
    "classtypes", "dataclass", "functypes", "methtypes"
]



"""
A series of decorators for use with and in *classes* only.
These functions are meant to provide both EnforceTypes wrapping, as well as making
sure that the handling order for common combined decorators is safely handled.
Reduces combined decorators down to one by applying decorators like ``@dataclass``
and ``@staticmethod`` in the right order.
"""

from functools import wraps
from inspect import _ParameterKind as PK
from inspect import signature
from types import FunctionType
from typing import Any, Callable, Type, cast
from typing_extensions import get_args

from .types import MISSING, F, T


def classtypes(cls: Type[T]) -> Type[T]:
    """For use as class decorator. Enforces types in ``init``"""
    oldinit: Callable = getattr(cls, "__init__")
    # If this constructor is a built-in object or otherwise a wrapper around something
    # then instantiation behavior is different
    is_obj = not isinstance(oldinit, FunctionType)
    keywords = signature(cls).parameters

    def newinit(self: Type[T], *args: Any, **kwargs: Any) -> None:
        arglist = list(args)
        for kw in keywords:
            argval = arglist.pop(0) if len(arglist) > 0 else ...
            kw_val = kwargs.get(kw, MISSING)
            if argval is not ... and kw_val is not MISSING:
                raise TypeError(f"{kw} was given as both a positional and a keyword argument! "
                                f"({cls.__name__})")
            if kw_val is MISSING:
                if argval is ...:
                    continue
                kw_val = argval
                if keywords[kw].kind == PK.POSITIONAL_ONLY:
                    arglist.append(kw_val)
                else:
                    if keywords[kw].kind == PK.KEYWORD_ONLY and kw not in kwargs:
                        raise SyntaxError(f"{kw} is a keyword-only argument, but was given as "
                                          f"positional argument! ({cls.__name__})")
                    kwargs[kw] = kw_val
            argtype = keywords[kw].annotation
            argtuple = get_args(argtype)
            if not isinstance(kw_val, argtype) and not isinstance(kw_val, argtuple):
                raise TypeError(f"Argument {kw} was passed a value of type `"
                                f"{type(kw_val).__name__}`, but only accepts values of "
                                f"type `{argtype.__name__}` "
                                f"({cls.__name__})")
        if is_obj:
            oldinit(self)
        else:
            oldinit(self, **kwargs)

    setattr(cls, "__init__", newinit)
    return cls


def methtypes(meth: F) -> F:
    """For use as a decorator for methods and @classmethods where ``self`` or ``cls`` is passed"""
    assert meth

    @wraps(meth)
    def newcall(selfcls: Type[T], *args: Any, **kwargs: Any) -> Any:
        keywords = signature(meth).parameters
        arglist = list(args)
        for kw in keywords:
            if kw in ["self", "cls"]:
                continue
            argval = arglist.pop(0) if len(arglist) > 0 else ...
            kw_val = kwargs.get(kw, MISSING)
            if argval is not ... and kw_val is not MISSING:
                raise SyntaxError(f"{kw} was given as both a positional and a keyword argument! "
                                  f"({meth.__name__})")
            if kw_val is MISSING:
                if argval is ...:
                    continue
                kw_val = argval
                if keywords[kw].kind == PK.POSITIONAL_ONLY:
                    arglist.append(kw_val)
                else:
                    if keywords[kw].kind == PK.KEYWORD_ONLY and kw not in kwargs:
                        raise SyntaxError(f"{kw} is a keyword-only argument, but was given as "
                                          f"positional argument! ({meth.__name__})")
                    kwargs[kw] = kw_val
            argtype = keywords[kw].annotation
            argtuple = get_args(argtype)
            if not isinstance(kw_val, argtype) and not isinstance(kw_val, argtuple):
                raise TypeError(f"Argument {kw} was passed a value of type `"
                                f"{type(kw_val).__name__}`, but only accepts values of "
                                f"type `{argtype.__name__}` "
                                f"({meth.__name__})")
        return meth(selfcls, *arglist, **kwargs)

    return cast(F, newcall)
